#! /usr/bin/python2
# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is bezier.py
#
# The Initial Developer of the Original Code is
# Robert Mark Waugh, bezier at waughonline dot com
# Portions created by the Initial Developer are Copyright (C) 2003
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# ***** END LICENSE BLOCK *****

# bezier.py
# escape quits, mouse button pauses/unpauses, 'e' shows 
# A bezier curve implementation that will pick a random number of control
# points and calculate a bezier curve within their convex hull.  
#
# Bezier curves are an integral part of defining curvelinear motion in 
# computing, derived originally in the 1960's by Bezier, an engineer 
# at Renault, to solve the problem of representing arbitrarily curved 
# surfaces in a minimal amount of space.  They have later been applied
# heavily to fonts and other graphical endevours.
#
# A bezier curve of four points can describe most curves, and more
# complex curves are generally fashioned by chaining together these cubic
# bezier curves one after another, as the nCr function for higher degree
# bezier curves becomes hard to compute.  There are some major limitations
# to a bezier curve.  A bezier curve can not represent a circle.  The
# parallel to a bezier curve can not be easily identified, except in the
# case of the same curved shifted.  The control points do not lay directly
# on the curve so it is sometimes non-intuitive what will occur.  A bezier
# curve of one point is a point, a bezier curve of two points is a line,
# and a bezier curve of three points where P1 < P2 < P3 is a quadratic, etc.
# The derivative of the first point of the curve is the slope between P1
# and P2, as is the derivative of the final point the slope between
# Pn and Pn-1.  The bezier point connects to only two points, the first
# and final points.
#
# This particular code sample will provide a screen saver like application
# which will visibly show you the path the curve is taking.  When it slows
# down the curve itself is slowing down, ie, the distance between computed
# points has decreased considerably.  This is intentional to give insight
# into the nature of bezier curves for the purpose of calculating curved
# paths.
#
# As can be guessed, there is particular relevence between bezier curves and
# computer game generation.  Bezier curves can concisely describe paths of
# movement for sprites across a surface.  These paths can be random (as
# demonstrated in this application) or can be pregenerated and recorded
# as a series of points.  This allows the game designer to be freed from
# creating canned seeming paths that are entirely linear or are variations
# on sinusoidal functions, and can appear more "natural."  They can also
# be used to simulate handwriting, signatures, etc.  It is also useful
# for animation sequences that need non-linear time sequencing ... by
# spacing the bezier points out in a straight line and manipulating their
# relative positions you can speed up and slow down the animations.
#
# The slowness of this implementation is largely due to the animation of
# the curve and not the computation of the points themselves, although
# computing the original equation of curves of very high degrees become
# significantly slower.  However, don't worry about it because you can't
# compute a binomial coefficient of a greater degree than 30 with this
# particular implementation. This implementation is faster than most
# since it computes the factorial parts up front instead of recomputing
# every time.

import os,sys, time, random
import pygame
from pygame.locals import *
import bezier_curve

clock = pygame.time.Clock()
pygame.init()
default_mode = pygame.display.list_modes()[0]
screen = pygame.display.set_mode(default_mode, FULLSCREEN)
pygame.mouse.set_visible(0)
pause = 0
equation = 0

def rest(seconds):
	handle_events()
	pygame.display.flip()
	time.sleep(seconds)

def handle_events():
	global pause, equation
	events = pygame.event.get()
	for event in events:
		if event.type == MOUSEBUTTONDOWN:
			pause ^= 1
			while pause:
				rest(.1)
		if event.type == KEYDOWN:
			if event.key == K_ESCAPE:
				sys.exit()
			elif event.key == 101:
				equation ^= 1

def create_bezier_curve(screen):
	points = []
	num_points = rand_factor(30);
	num_points = 30
	if num_points < 3:
		num_points += 2
	for i in range(0, num_points):
		points.append(rand_dim())

	background = pygame.Surface(screen.get_size())
	background.fill((0, 0, 0))
	curve = bezier_curve.BezierCurve(points)
	if equation:
		render_equation(curve, background)
	render_points(points, background)
	render_bezier(curve, 0.007, background)

def render_equation(curve, background):
	font = pygame.font.Font(None, 20)
	terms = curve.get_as_string().split("+")
	position = ( 5, 5 )
	index = 0
	for term in terms:
		if index:
			term = " + " + term
		term_text = font.render(term, 1, (200, 200, 200))
		text_size = font.size(term)
		if text_size[0]+position[0] >= background.get_rect().width:
			vertical_shift = \
				       position[1]+text_size[1]+\
				       font.get_linesize()
			position = ( 10, vertical_shift )
		position = (position[0]+text_size[0], position[1])
		text_pos = term_text.get_rect()
		text_pos.topright = position
		background.blit(term_text, text_pos)
		index += 1
	return background

def render_bezier(curve, factor, background):
	previous_point = 0
	for point in curve.get_curve(factor):
		if previous_point:
			connect_points(previous_point, point, background)
			pygame.display.flip()
		previous_point = point
		handle_events()
	control_points = curve.get_points()
	connect_points(previous_point, control_points[len(control_points)-1],\
		 background)
	pygame.display.flip()

	
def render_points(control_points, background):
	rects = []
	for point in control_points:
		new_rect = pygame.Rect(((0, 0), (5, 5)))
		new_rect.centerx = point[0]
		new_rect.centery = point[1]
		rects.append(new_rect)
	pygame.draw.lines(background, (0, 255, 0), 0, control_points, 1)
	for rect in rects:
		pygame.draw.ellipse(background, (255, 0, 0), rect, 0)
	screen.blit(background, background.get_rect())

def connect_points(previous_point, point, background):
	pygame.draw.line(background, (255, 255, 255),\
		previous_point, point, 3)
	screen.blit(background, background.get_rect())

def rand_factor(factor):
	return random.random()*factor

def rand_dim():
	x = rand_factor(default_mode[0])
	if equation:
		y = rand_factor(default_mode[1]-143)+143
	else:
		y = rand_factor(default_mode[1])
	return (x,y)

while 1:
	if not pause:
		create_bezier_curve(screen)
	for i in range(0, 50):
		rest(.1)
