# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is bezier_curve.py
#
# The Initial Developer of the Original Code is
# Robert Mark Waugh, bezier at waughonline dot com
# Portions created by the Initial Developer are Copyright (C) 2003
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# ***** END LICENSE BLOCK *****

class BezierCurve:
    def __init__(self, control_points):
        """ pre-compute the coefficients to save time when rendering the
            curve """
        self.control_points = control_points
        self.bezier = []
        n = len(control_points) - 1
        for k in range(0, (n+1)):
            coef = self.nCr(n, k)
            self.bezier.append({ "x": control_points[k][0],\
                                 "y": control_points[k][1],\
                                 "coef": coef, "n":n, "k":k})

    def get_points(self):
        return self.control_points

    def get_curve(self, factor):
        """ The factor is a fraction of 1, which is a stepping factor.
            ie, a factor of .01 will give you 100 points of the curve. """
	current = 0
	coordinates = []
	while current < 1:
		coordinates.append(self.calculate_coord(current))
		current += factor
	return coordinates

    def calculate_coord(self,u):
        """ http://mathworld.wolfram.com/BezierCurve.html """
        coord = [0, 0]
        for i in range(0,  len(self.bezier)):
            term = self.bezier[i]
            p_x = term["x"]
            p_y = term["y"]
            coef = term["coef"]
            n = term["n"]
            k = term["k"]
            coord[0] += coef * p_x * (u**k)*(1-u)**(n-k)
            coord[1] += coef * p_y * (u**k)*(1-u)**(n-k)
        return (coord[0], coord[1])

    def get_as_string(self):
        """ returns the bezier curve as a math function """
        strng = "C(u) = "
        for i in range(0, len(self.bezier)):
            
           if i:
               strng += " + "
           strng += str(self.bezier[i]["coef"]) + "(P" + str(i) +\
                  ")(u^"+str(self.bezier[i]["k"])+")(1-u)^("+\
                  str(self.bezier[i]["n"]-self.bezier[i]["k"])+")"
        return strng

    def nCr(self, n, r):
        """ http://mathworld.wolfram.com/BinomialCoefficient.html """
	if n == r or r == 0:
		return 1
	total = 1
	n_fact = n
	r_fact = 1
	nr_fact = 1
	while n_fact != 0:
		total = total * n_fact
		n_fact -= 1
		if r_fact < total and r_fact <= r:
			total = total / r_fact
			r_fact += 1
		if nr_fact < total and nr_fact <= (n - r):
			total = total / nr_fact
			nr_fact += 1
	return total

