#! /usr/bin/python2
# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is bezier.py
#
# The Initial Developer of the Original Code is
# Robert Mark Waugh, bezier at waughonline dot com
# Portions created by the Initial Developer are Copyright (C) 2003
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# ***** END LICENSE BLOCK *****

# bezier.py
# escape quits, mouse button pauses/unpauses
# A bezier curve implementation that will pick a random number of control
# points and calculate a bezier curve within their convex hull.  
#
# Bezier curves are an integral part of defining curvelinear motion in 
# computing, derived originally in the 1960's by Bezier, an engineer 
# at Renault, to solve the problem of representing arbitrarily curved 
# surfaces in a minimal amount of space.  They have later been applied
# heavily to fonts and other graphical endevours.
#
# A bezier curve of four points can describe most curves, and more
# complex curves are generally fashioned by chaining together these cubic
# bezier curves one after another, as the nCr function for higher degree
# bezier curves becomes hard to compute.  There are some major limitations
# to a bezier curve.  A bezier curve can not represent a circle.  The
# parallel to a bezier curve can not be easily identified, except in the
# case of the same curved shifted.  The control points do not lay directly
# on the curve so it is sometimes non-intuitive what will occur.  A bezier
# curve of one point is a point, a bezier curve of two points is a line,
# and a bezier curve of three points where P1 < P2 < P3 is a quadratic, etc.
# The derivative of the first point of the curve is the slope between P1
# and P2, as is the derivative of the final point the slope between
# Pn and Pn-1.  The bezier point connects to only two points, the first
# and final points.
#
# This particular code sample will provide a screen saver like application
# which will visibly show you the path the curve is taking.  When it slows
# down the curve itself is slowing down, ie, the distance between computed
# points has decreased considerably.  This is intentional to give insight
# into the nature of bezier curves for the purpose of calculating curved
# paths.
#
# As can be guessed, there is particular relevence between bezier curves and
# computer game generation.  Bezier curves can concisely describe paths of
# movement for sprites across a surface.  These paths can be random (as
# demonstrated in this application) or can be pregenerated and recorded
# as a series of points.  This allows the game designer to be freed from
# creating canned seeming paths that are entirely linear or are variations
# on sinusoidal functions, and can appear more "natural."  They can also
# be used to simulate handwriting, signatures, etc.  The slowness of this
# implementation is largely due to the animation of the curve and not the
# computation of the points themselves, although curves of high degrees
# become significantly slower, with what I believe is an O(n!) running time.
# For degrees of less than 10 it seems to be trivial, but above that it
# becomes significant.  Also, you can not compute binomial coefficients
# of a greater degree than 35 with this particular implementation.  I have
# thought of some ways to do that but it doesn't seem useful.  If others out
# there have a good efficient nCr function that can compute arbitrarily large
# coefficients I would love to see it.
#
# This code is not very modularized and is mainly here to showcase these 
# functions:
# render_points(list = [(x,y),...])
# calculate_points(list = [(x,y),...], u)
# calculate_coord(list = [(x,y),...])
# nCr(n, r)
#

import os,sys, time, random
import pygame
from pygame.locals import *


clock = pygame.time.Clock()
pygame.init()
default_mode = pygame.display.list_modes()[0]
screen = pygame.display.set_mode(default_mode, FULLSCREEN)
pygame.mouse.set_visible(0)
pause = 0
font = pygame.font.Font(None, 36)
text = font.render("Reticulating splines...", 1, (0, 0, 255))
text.get_rect().centerx = screen.get_rect().centerx

def rest(seconds):
	handle_events()
	pygame.display.flip()
	time.sleep(seconds)

def handle_events():
	global pause
	events = pygame.event.get()
	for event in events:
		if event.type == MOUSEBUTTONDOWN:
			pause ^= 1
			while pause:
				rest(.1)
		if event.type == KEYDOWN:
			if event.key == K_ESCAPE:
				sys.exit()

def create_bezier_curve(screen):
	points = []
	num_points = rand_dim(30);
	if num_points < 3:
		num_points += 2
	for i in range(0, num_points):
		points.append((rand_dim(default_mode[0]), \
			rand_dim(default_mode[1])))
	

	render_points(points)

def render_points(control_points):
	rects = []
	for point in control_points:
		new_rect = pygame.Rect(((0, 0), (5, 5)))
		new_rect.centerx = point[0]
		new_rect.centery = point[1]
		rects.append(new_rect)
	background = pygame.Surface(screen.get_size())
	background.fill((0, 0, 0))
	pygame.draw.lines(background, (0, 255, 0), 0, control_points, 1)
	for rect in rects:
		pygame.draw.ellipse(background, (255, 0, 0), rect, 0)
	screen.blit(background, background.get_rect())
	curve_points = calculate_points(control_points, .007)
	previous_point = 0
	for point in curve_points:
		if previous_point:
			connect_points(previous_point, point, background)
			pygame.display.flip()
		previous_point = point
		handle_events()
	connect_points(previous_point, control_points[len(control_points)-1],\
		 background)
	pygame.display.flip()

def connect_points(previous_point, point, background):
	pygame.draw.line(background, (255, 255, 255),\
		previous_point, point, 3)
	screen.blit(background, background.get_rect())

def rand_dim(factor):
	return random.random()*factor

def calculate_points(points, factor):
	current = 0
	coordinates = []
	while current < 1:
		coordinates.append(calculate_coord(points, current))
		current += factor

	return coordinates

def calculate_coord(points, current):
	n = len(points) - 1
	coord = [0,0]
	# the X and Y values are derived by performing a progression of
	# Bernstein polynomials multiplied by the x/y magnitude of P[k].
	# http://mathworld.wolfram.com/BezierCurve.html
	for k in range(0, len(points)):
		coord[0] = coord[0] + points[k][0] * \
			(nCr(n,k)*(current**k)*((1-current)**(n-k)))
		coord[1] = coord[1] + points[k][1] * \
			(nCr(n,k)*(current**k)*((1-current)**(n-k)))
	return coord

def nCr(n, r):
	if n == r or r == 0:
		return 1
	total = 1
	n_fact = n
	r_fact = 1
	nr_fact = 1
	# the numerator is calculated from the right, but the denominator
	# is calculated from the left.  This allows us to compute larger
	# numbers without seeing overflows, but doesn't introduce the same
	# level of rounding problems as approaching both from the right
	# or left.
	# Another way of doing this better would be to start in the middle
	# and work outwards.
	while n_fact != 0:
		total = total * n_fact
		n_fact -= 1
		if r_fact < total and r_fact <= r:
			total = total / r_fact
			r_fact += 1
		if nr_fact < total and nr_fact <= (n - r):
			total = total / nr_fact
			nr_fact += 1
	return total


while 1:
	if not pause:
		screen.blit(text, text.get_rect())
		pygame.display.flip()
		create_bezier_curve(screen)
	for i in range(0, 50):
		rest(.1)
